@@ -2,10 +2,13 @@ import json
 import os
 import os.path as op
 from pathlib import Path
+import re
 from shutil import rmtree
 
 import pytest
+import responses
 
+from .skip import mark
 from ..consts import DRAFT
 from ..download import download
 from ..utils import find_files
@@ -13,13 +16,11 @@ from ..utils import find_files
 
 # both urls point to 000027 (lean test dataset), and both draft and "released"
 # version have only a single file ATM
+@mark.skipif_no_network
 @pytest.mark.parametrize(
     "url",
     [  # Should go through API
-        pytest.param(
-            "https://dandiarchive.org/dandiset/000027/0.200721.2222",
-            marks=pytest.mark.xfail(reason="publish.dandiarchive.org is gone"),
-        ),
+        "https://dandiarchive.org/dandiset/000027/0.210831.2033",
         # Drafts do not go through API ATM, but that should not be visible to user
         "https://dandiarchive.org/dandiset/000027/draft",
     ],
@@ -51,13 +52,11 @@ def test_download_000027(url, tmpdir):
     download(url, tmpdir, existing="refresh")  # TODO: check that skipped (the same)
 
 
+@mark.skipif_no_network
 @pytest.mark.parametrize(
     "url",
     [  # Should go through API
-        pytest.param(
-            "https://dandiarchive.org/dandiset/000027/0.200721.2222",
-            marks=pytest.mark.xfail(reason="publish.dandiarchive.org is gone"),
-        ),
+        "https://dandiarchive.org/dandiset/000027/0.210831.2033",
         # Drafts do not go through API ATM, but that should not be visible to user
         "https://dandiarchive.org/dandiset/000027/draft",
     ],
@@ -70,13 +69,11 @@ def test_download_000027_metadata_only(url, tmpdir):
     assert sorted(downloads) == ["dandiset.yaml"]
 
 
+@mark.skipif_no_network
 @pytest.mark.parametrize(
     "url",
     [  # Should go through API
-        pytest.param(
-            "https://dandiarchive.org/dandiset/000027/0.200721.2222",
-            marks=pytest.mark.xfail(reason="publish.dandiarchive.org is gone"),
-        ),
+        "https://dandiarchive.org/dandiset/000027/0.210831.2033",
         # Drafts do not go through API ATM, but that should not be visible to user
         "https://dandiarchive.org/dandiset/000027/draft",
     ],
@@ -89,17 +86,9 @@ def test_download_000027_assets_only(url, tmpdir):
     assert sorted(downloads) == ["sub-RAT123", op.join("sub-RAT123", "sub-RAT123.nwb")]
 
 
+@mark.skipif_no_network
 @pytest.mark.parametrize("resizer", [lambda sz: 0, lambda sz: sz // 2, lambda sz: sz])
-@pytest.mark.parametrize(
-    "version",
-    [
-        pytest.param(
-            "0.200721.2222",
-            marks=pytest.mark.xfail(reason="publish.dandiarchive.org is gone"),
-        ),
-        DRAFT,
-    ],
-)
+@pytest.mark.parametrize("version", ["0.210831.2033", DRAFT])
 def test_download_000027_resume(tmp_path, resizer, version):
     from ..support.digests import Digester
 
@@ -132,6 +121,7 @@ def test_download_newest_version(local_dandi_api, text_dandiset, tmp_path):
     dandiset_id = text_dandiset["dandiset_id"]
     download(f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}", tmp_path)
     assert (tmp_path / dandiset_id / "file.txt").read_text() == "This is test text.\n"
+    text_dandiset["dandiset"].wait_until_valid()
     text_dandiset["dandiset"].publish()
     (text_dandiset["dspath"] / "file.txt").write_text("This is different text.\n")
     text_dandiset["reupload"]()
@@ -245,3 +235,16 @@ def test_download_sync_list(capsys, local_dandi_api, mocker, text_dandiset, tmp_
         mocker.call("Delete 1 local asset? ([y]es/[n]o/[l]ist): "),
     ]
     assert capsys.readouterr().out.splitlines()[-1] == str(dspath / "file.txt")
+
+
+@responses.activate
+def test_download_no_blobDateModified(local_dandi_api, text_dandiset, tmp_path):
+    # Regression test for #806
+    responses.add_passthru(re.compile("^http"))
+    client = text_dandiset["client"]
+    dandiset = text_dandiset["dandiset"]
+    asset = dandiset.get_asset_by_path("file.txt")
+    metadata = asset.get_raw_metadata()
+    del metadata["blobDateModified"]
+    responses.add(responses.GET, client.get_url(asset.api_path), json=metadata)
+    download(client.get_url(dandiset.api_path), tmp_path)
